#
# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
# Copyright (C) 2016-2021 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -*- coding: utf-8 -*-

import setuptools

setuptools.setup(
    name="managesf",
    setup_requires=['pbr'],
    packages=["managesf"],
    description="REST API interface and CLI for software-factory project",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    pbr=True)
